import numpy as np
from scipy.stats import t


def pooled_standard_deviation(sample1: np.ndarray, sample2: np.ndarray) -> float:
    """
    Compute the pooled standard deviation for two samples.

    Args:
        sample1 (np.ndarray): The first sample as a 1D NumPy array.
        sample2 (np.ndarray): The second sample as a 1D NumPy array.

    Returns:
        float: The pooled standard deviation.
    """
    var1 = np.var(sample1, ddof=1)
    var2 = np.var(sample2, ddof=1)
    n1 = len(sample1)
    n2 = len(sample2)

    s_pooled = np.sqrt(((n1 - 1) * var1 + (n2 - 1) * var2) / (n1 + n2 - 2))

    return s_pooled


def two_sample_t_test_pooled(sample1: np.ndarray, sample2: np.ndarray) -> float:
    """
    Compute the independent two-sample t-test statistic using pooled standard deviation.

    Args:
        sample1 (np.ndarray): The first sample as a 1D NumPy array.
        sample2 (np.ndarray): The second sample as a 1D NumPy array.

    Returns:
        float: The t-test statistic.
    """
    # Calculate the means of the two samples
    mean1 = np.mean(sample1)
    mean2 = np.mean(sample2)
    n1 = len(sample1)
    n2 = len(sample2)
    # Calculate the pooled standard deviation
    s_pooled = pooled_standard_deviation(sample1, sample2)
    t = (mean1 - mean2) / (s_pooled * np.sqrt(1 / n1 + 1 / n2))
    return t


def cohens_d_from_data(sample1: np.ndarray, sample2: np.ndarray) -> float:
    """
    Compute Cohen's d effect size for two samples.

    Args:
        sample1 (np.ndarray): The first sample as a 1D NumPy array.
        sample2 (np.ndarray): The second sample as a 1D NumPy array.

    Returns:
        float: The Cohen's d effect size.
    """
    # Calculate the means of the two samples
    mean1 = np.mean(sample1)
    mean2 = np.mean(sample2)
    # Calculate the pooled standard deviation
    s_pooled = pooled_standard_deviation(sample1, sample2)
    # Calculate Cohen's d effect size
    d = (mean1 - mean2) / s_pooled
    return d


def cohens_d_from_statistics(
    mean1: float, mean2: float, std1: float, std2: float, n1: int, n2: int
) -> float:
    """
    Compute Cohen's d effect size for two samples given their means and standard
    deviations.

    Args:
        mean1 (float): The mean of the first sample.
        mean2 (float): The mean of the second sample.
        std1 (float): The standard deviation of the first sample.
        std2 (float): The standard deviation of the second sample.
        n1 (int): The size of the first sample.
        n2 (int): The size of the second sample.

    Returns:
        float: The Cohen's d effect size.
    """
    # Calculate the pooled standard deviation
    s_pooled = np.sqrt(((n1 - 1) * std1**2 + (n2 - 1) * std2**2) / (n1 + n2 - 2))
    # Calculate Cohen's d effect size
    d = (mean1 - mean2) / s_pooled
    return d


def hedges_g(sample1: np.ndarray, sample2: np.ndarray) -> float:
    """
    Compute Hedge's g as a measure of effect size for two samples.

    Args:
        sample1 (np.ndarray): The first sample as a 1D NumPy array.
        sample2 (np.ndarray): The second sample as a 1D NumPy array.

    Returns:
        float: The Hedge's g value.
    """
    n1 = len(sample1)
    n2 = len(sample2)
    g = cohens_d_from_data(sample1, sample2) * (1 - (3 / (4 * (n1 + n2) - 9)))
    return g


def cohens_d_from_mde_absolute(mde_absolute: float, pooled_std_dev: float) -> float:
    """
    Compute the Cohen's d effect size given a minimal detectable effect
    (MDE) in terms of an absolute difference in means and the pooled standard deviation
    of the groups.

    Args:
        mde_absolute (float): The minimal detectable effect in terms of an absolute.
                              difference in means.
        pooled_std_dev (float): The pooled standard deviation of the groups.

    Returns:
        float: The Cohen's d effect size.
    """
    return mde_absolute / pooled_std_dev


def cohens_d_from_mde_relative(
    sample1_mean: float, mde_relative: float, pooled_std: float
) -> float:
    """
    Compute the Cohen's d effect size given a minimal detectable effect
    (MDE) in terms of an percentage difference in means and the pooled standard deviation
    of the groups.

    Args:
        mde_relative (float): The minimal detectable effect in terms of a percentage.
                              difference in means.
        pooled_std (float): The pooled standard deviation of the groups.

    Returns:
        float: The Cohen's d effect size.
    """
    return cohens_d_from_mde_absolute(sample1_mean / 100 * mde_relative, pooled_std)


def compute_bootstrap_ci_cohens_d(
    sample1, sample2, num_bootstrap_samples=10000, ci=0.95
):
    """
    Compute a confidence interval around Cohen's d effect size using bootstrap
    resampling (for small sample size).

    Args:
        sample1 (np.array): The first group of samples.
        sample2 (np.array): The second group of samples.
        num_bootstrap_samples (int): The number of bootstrap samples to generate.
        Default is 10000.
        ci (float): The desired confidence level for the interval (e.g., 0.95
        for a 95% CI). Default is 0.95.

    Returns:
        tuple: The lower and upper bounds of the confidence interval.
    """
    # Generate bootstrap samples and compute Cohen's d for each sample
    bootstrap_samples = []
    for _ in range(num_bootstrap_samples):
        bootstrap_sample1 = np.random.choice(sample1, size=len(sample1), replace=True)
        bootstrap_sample2 = np.random.choice(sample2, size=len(sample2), replace=True)
        bootstrap_samples.append(
            cohens_d_from_data(bootstrap_sample1, bootstrap_sample2)
        )

    # Compute the percentiles of the bootstrap samples to get the confidence
    # interval
    lower_bound = np.percentile(bootstrap_samples, (1 - ci) / 2 * 100)
    upper_bound = np.percentile(bootstrap_samples, (1 + ci) / 2 * 100)

    return lower_bound, upper_bound


def compute_parametric_ci_cohens_d(sample1, sample2, ci=0.95):
    """
    Compute a confidence interval around Cohen's d effect size using a parametric
    method (for large sample size).

    Args:
        sample1 (np.array): The first group of samples.
        sample2 (np.array): The second group of samples.
        ci (float): The desired confidence level for the interval (e.g., 0.95
        for a 95% CI). Default is 0.95.

    Returns:
        tuple: The lower and upper bounds of the confidence interval.
    """
    # Calculate the parameters
    n1, n2 = len(sample1), len(sample2)
    d = cohens_d_from_data(sample1, sample2)

    # Compute the standard error of Cohen's d
    std_error_d = np.sqrt((n1 + n2) / (n1 * n2) + d**2 / (2 * (n1 + n2)))

    # Compute the confidence interval for Cohen's d
    df = n1 + n2 - 2  # degrees of freedom
    t_critical = t.ppf((1 + ci) / 2, df)
    ci_lower_d = d - t_critical * std_error_d
    ci_upper_d = d + t_critical * std_error_d

    return ci_lower_d, ci_upper_d



import numpy as np
import statsmodels.stats.power as smp
from statsmodels.stats.weightstats import ttest_ind

from .effect import (
    cohens_d_from_mde_absolute,
    cohens_d_from_mde_relative,
    compute_bootstrap_ci_cohens_d,
    compute_parametric_ci_cohens_d,
    pooled_standard_deviation,
)


class TTestInd:
    """
    Class to perform Independent T-Test and related statistics.

    Attributes:
        control (np.ndarray): Sample data for control group.
        treatment (np.ndarray): Sample data for treatment group.
        mde_absolute (float | None): Minimum detectable effect in absolute terms.
        mde_relative (float | None): Minimum detectable effect in relative terms.
        control_n (int): Number of observations in control group.
        treatment_n (int): Number of observations in treatment group.
        ratio (float): Ratio of number of observations in treatment group to control group.
        control_mean (float): Mean of control group.
        treatment_mean (float): Mean of treatment group.
        control_std (float): Standard deviation of control group.
        treatment_std (float): Standard deviation of treatment group.
        pooled_std (float): Pooled standard deviation of control and treatment group.
        effect_size (float): Calculated effect size.
        power (float | None): Calculated test power.
        t_stat (float | None): Calculated t-statistic.
        p_value (float | None): Calculated p-value.
        ci_lower (float | None): Lower limit of confidence interval.
        ci_upper (float | None): Upper limit of confidence interval.
    """

    control: np.ndarray
    treatment: np.ndarray

    mde_absolute: float | None
    mde_relative: float | None

    control_n: int
    treatment_n: int
    ratio: float
    control_mean: float
    treatment_mean: float
    control_std: float
    treatment_std: float
    pooled_std: float
    effect_size: float

    power: float | None = None
    t_stat: float | None = None
    p_value: float | None = None
    ci_lower: float | None = None
    ci_upper: float | None = None

    def __init__(
        self,
        control: np.ndarray,
        treatment: np.ndarray,
        mde_absolute: float | None = None,
        mde_relative: float | None = None,
    ):
        """
        Initialize TTestInd with control and treatment data, and either of
        mde_absolute or mde_relative.

        Args:
            control (np.ndarray): Sample data for control group.
            treatment (np.ndarray): Sample data for treatment group.
            mde_absolute (float | None): Minimum detectable effect in absolute
            terms.
            mde_relative (float | None): Minimum detectable effect in relative
            terms.
        """
        if not mde_absolute and not mde_relative:
            raise ValueError("Either mde_absolute or mde_relative must be provided.")
        if mde_absolute and mde_relative:
            raise ValueError(
                "Only one of mde_absolute or mde_relative can be provided."
            )
        self.control = control
        self.treatment = treatment
        self.mde_absolute = mde_absolute
        self.mde_relative = mde_relative
        self.compute_samples_statistics()
        self.compute_effect_size()

    def compute_samples_statistics(self) -> None:
        """
        Compute basic statistics for control and treatment data.
        """
        self.control_n = len(self.control)
        self.treatment_n = len(self.treatment)
        self.ratio = self.treatment_n / self.control_n
        self.control_mean = np.mean(self.control)
        self.treatment_mean = np.mean(self.treatment)
        self.control_std = np.std(self.control)
        self.treatment_std = np.std(self.treatment)
        self.pooled_std = pooled_standard_deviation(self.control, self.treatment)

    def compute_effect_size(self) -> float:
        """
        Compute the effect size based on the provided minimum detectable effect.

        Returns:
            float: The computed effect size.
        """
        if self.mde_absolute:
            self.effect_size = cohens_d_from_mde_absolute(
                self.mde_absolute, self.pooled_std
            )
        else:
            self.effect_size = cohens_d_from_mde_relative(
                self.control_mean, self.mde_relative, self.pooled_std
            )
        return self.effect_size

    def conduct_test(self, alternative: str = "two-sided") -> tuple[float, float]:
        """
        Conduct the independent t-test.

        Args:
            alternative (str, optional): The alternative hypothesis, can be
            "two-sided", "larger" or "smaller". Defaults to "two-sided".

        Returns:
            tuple[float, float]: The computed t-statistic and p-value.
        """
        self.t_stat, self.p_value, _ = ttest_ind(
            self.control, self.treatment, alternative
        )
        return self.t_stat, self.p_value

    def compute_confidence_interval(
        self, percentage: float = 0.95
    ) -> tuple[float, float]:
        """
        Compute the confidence interval for the effect size.

        Args:
            percentage (float, optional): The confidence level. Defaults to 0.95.

        Returns:
            tuple[float, float]: The lower and upper limit of the confidence
            interval.
        """
        if self.control_n > 30 and self.treatment_n > 30:
            self.ci_lower, self.ci_upper = compute_parametric_ci_cohens_d(
                self.control, self.treatment, percentage
            )
        else:
            self.ci_lower, self.ci_upper = compute_bootstrap_ci_cohens_d(
                self.control, self.treatment, percentage
            )
        return self.ci_lower, self.ci_upper

    def compute_power(self, alpha: float = 0.05) -> float:
        """
        Compute the power of the test.

        Args:
            alpha (float, optional): The significance level. Defaults to 0.05.

        Returns:
            float: The computed power.
        """
        self.power = smp.tt_ind_solve_power(
            self.effect_size,
            power=None,
            nobs1=self.control_n,
            ratio=self.ratio,
            alpha=alpha,
        )
        return self.power

    def report(self) -> dict[str, int | float | str]:
        """
        Generate a report of the analysis.

        Returns:
            dict[str, int | float | str]]: A dictionary containing the number of
            samples, means, standard deviations for control and treatment groups,
            minimum detectable effects (absolute and relative), power of the
            test, t-statistic, p-value, effect size (Cohen's d), and the
            confidence interval for the effect size.
        """
        return {
            "control_n": self.control_n,
            "control_mean": round(self.control_mean, 3),
            "control_std": round(self.control_std, 3),
            "treatment_n": self.treatment_n,
            "treatment_mean": round(self.treatment_mean, 3),
            "treatment_std": round(self.treatment_std, 3),
            "mde_absolute": self.mde_absolute,
            "mde_relative": self.mde_relative,
            "power": round(self.power, 3),
            "t_stat": round(self.t_stat, 3),
            "p_value": round(self.p_value, 3),
            "cohens_d": round(self.effect_size, 3),
            "CI": f"[{round(self.ci_lower, 3)}, {round(self.ci_upper, 3)}]",
        }



"""GeoDistance module for calculating road distances between two points on Earth."""
import time
import warnings
from math import asin, cos, radians, sin, sqrt

import openrouteservice
from openrouteservice.exceptions import ApiError


class GeoDistance:
    """GeoDistance class for calculating road distances between two points on Earth.

    Attributes:
        api_keys (list[str]): List of API keys for the OpenRouteService API.
        client (openrouteservice.Client): Client for the OpenRouteService API.

    Methods:
        get_geodistance(origin_latitude: float, origin_longitude: float,
        destination_latitude: float, destination_longitude: float) -> float:

    """

    api_keys: list[str]
    client: openrouteservice.Client

    def __init__(self, api_keys: list[str] = []) -> None:
        """Initialize GeoDistance class.

        Args:
            api_keys (list[str]): List of API keys for the OpenRouteService API.
            Defaults to [].
        """
        if api_keys:
            self.api_keys = api_keys
            self.client = self._initialize_client(self.api_keys[0])

    @staticmethod
    def _initialize_client(api_key: str) -> openrouteservice.Client:
        """Initialize OpenRouteService client.

        Args:
            api_key (str): API key for the OpenRouteService API.
        """
        client = openrouteservice.Client(api_key)
        return client

    def get_openroute_distance(
        self,
        origin_latitude: float,
        origin_longitude: float,
        destination_latitude: float,
        destination_longitude: float,
        null_value: any = None,
    ) -> tuple[float, float]:
        """Calculate road distance between two points on Earth.

        Args:
            origin_latitude (float): Latitude of origin point.
            origin_longitude (float): Longitude of origin point.
            destination_latitude (float): Latitude of destination point.
            destination_longitude (float): Longitude of destination point.
            null_value (any, optional): Value to return if distance cannot be calculated.

        Returns:
            tuple[float, float]: Tuple of distance in meters and duration in seconds.
        """
        coords: tuple[tuple[int]] = (
            (origin_longitude, origin_latitude),
            (destination_longitude, destination_latitude),
        )
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            try:
                response: dict[str, dict] = self.client.directions(coords)
            except (UserWarning, ApiError) as exception:
                print(exception)
                if len(self.api_keys) == 1:
                    raise IndexError("No more API keys available.") from exception
                self.api_keys.append(self.api_keys.pop(0))
                self.client = self._initialize_client(self.api_keys[0])
                time.sleep(2)
                return self.get_geodistance(
                    origin_latitude,
                    origin_longitude,
                    destination_latitude,
                    destination_longitude,
                )
        kilometers_distance: any = null_value
        minutes_distance: any = null_value
        try:
            kilometers_distance = response["routes"][0]["summary"]["distance"] / 1000
            minutes_distance = response["routes"][0]["summary"]["duration"] / 60
        except KeyError:
            pass
        return (kilometers_distance, minutes_distance)

    def get_harvesine_distance(
        self,
        origin_latitude: float,
        origin_longitude: float,
        destination_latitude: float,
        destination_longitude: float,
    ) -> float:
        """Calculate the distance in kilometers between two points on the earth.

        Args:
            origin_latitude (float): Latitude of origin point.
            origin_longitude (float): Longitude of origin point.
            destination_latitude (float): Latitude of destination point.
            destination_longitude (float): Longitude of destination point.

        Returns:
            float: Distance in kilometers.
        """
        lon1, lat1, lon2, lat2 = map(
            radians,
            [
                origin_longitude,
                origin_latitude,
                destination_longitude,
                destination_latitude,
            ],
        )
        dlon: float = lon2 - lon1
        dlat: float = lat2 - lat1
        a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
        c = 2 * asin(sqrt(a))
        return 6371 * c



"""Data shape of a DataFrame overview."""
import pandas as pd
from IPython.display import display


def get_shape(
    data_frame: pd.DataFrame, display_summary: bool = True
) -> dict[str, int | list[str]]:
    """Get a DataFrame shape overview.

    Args:
        data_frame (pd.DataFrame): DataFrame to get shape overview.

    Returns:
        dict[str, int | list[str]]: DataFrame shape summary.
    """
    summary: dict[str, int | list[str]] = {}
    summary["rows_number"], summary["columns_number"] = data_frame.shape
    summary["columns_name"] = data_frame.columns.tolist()
    summary["columns_type"] = data_frame.dtypes.tolist()
    if display_summary:
        display(data_frame.head())
        print(f"Number of rows: {summary.get('rows_number')}")
        print(f"Number of columns: {summary.get('columns_number')}")
        print(f"Columns name: {summary['columns_name']}")
        print(f"Columns type: {summary['columns_type']}")
    return summary



"""Exploratory Data Analysis (EDA) module for coolpandas."""
from .correlation import get_correlation
from .distribution import get_groupby_distribution
from .duplicates import duplicated_columns, duplicated_rows
from .features_type import (
    boolean_features,
    categorical_features,
    numerical_features,
    zero_variance_features,
)
from .geo_distance import GeoDistance
from .missing_values import get_missing_values
from .random_state import random_state
from .shape import get_shape
from .summary import get_summary
from .value_counts import get_groupby_value_counts, get_value_counts



"""Get features type in a DataFrame."""
import numpy as np
import pandas as pd


def categorical_features(
    data_frame: pd.DataFrame, display_summary: bool = True
) -> list[str]:
    """Get categorical features in a DataFrame.
    Args:
        data_frame (pd.DataFrame): DataFrame to get categorical features.
        display_summary (bool, optional): Whether to display summary. Defaults to True.

    Returns:
        list[str]: Categorical features.
    """
    features: list[str] = data_frame.select_dtypes(object).columns.tolist()
    if display_summary:
        print(f"Number of categorical features: {len(features)}")
        print(f"Categorical features: {features}")
    return features


def numerical_features(
    data_frame: pd.DataFrame, display_summary: bool = True
) -> list[str]:
    """Get numerical features in a DataFrame.
    Args:
        data_frame (pd.DataFrame): DataFrame to get numerical features.
        display_summary (bool, optional): Whether to display summary. Defaults to True.

    Returns:
        list[str]: Numerical features.
    """
    features: list[str] = data_frame.select_dtypes(include=np.number).columns.tolist()
    if display_summary:
        print(f"Number of numerical features: {len(features)}")
        print(f"Numerical features: {features}")
    return features


def boolean_features(
    data_frame: pd.DataFrame, display_summary: bool = True
) -> list[str]:
    """Get boolean features in a DataFrame.
    Args:
        data_frame (pd.DataFrame): DataFrame to get boolean features.
        display_summary (bool, optional): Whether to display summary. Defaults to True.

    Returns:
        list[str]: Boolean features.
    """
    features: list[str] = [
        i for i in data_frame.columns if data_frame[i].nunique() == 2
    ]
    if display_summary:
        print(f"Number of boolean features: {len(features)}")
        print(f"Boolean features: {features}")
    return features


def zero_variance_features(
    data_frame: pd.DataFrame, display_summary: bool = True
) -> list[str]:
    """Get zero variance features in a DataFrame.
    Args:
        data_frame (pd.DataFrame): DataFrame to get zero variance features.
        display_summary (bool, optional): Whether to display summary. Defaults to True.

    Returns:
        list[str]: Zero variance features.
    """
    features: list[str] = [
        i for i in data_frame.columns if data_frame[i].nunique() == 1
    ]
    if display_summary:
        print(f"Number of zero variance features: {len(features)}")
        print(f"Zero variance features: {features}")
    return features



"""Correlation map function."""
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from .style import custom_template, format_title


def mapplot(
    matrix: pd.DataFrame,
    title: str,
    subtitle: str | None = None,
    width: int = 1000,
    height: int = 1000,
    **kwargs,
) -> go.Figure:
    """Create a correlation map.

    Args:
        matrix (pd.DataFrame): Correlation matrix to plot.
        title (str): Title of the plot.
        subtitle (str, optional): Subtitle of the plot. Defaults to None.
        width (int, optional): Width of the plot. Defaults to 1000.
        height (int, optional): Height of the plot. Defaults to 1000.
        **kwargs: Keyword arguments to pass to plotly.express.imshow.

    Returns:
        go.Figure: Correlation map figure.
    """
    fig = px.imshow(
        matrix,
        text_auto=True,
        zmin=-1,
        zmax=1,
        title=format_title(title, subtitle=subtitle),
        template=custom_template,
        width=width,
        height=height,
        **kwargs,
    )
    return fig



"""Box plot module."""
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from .style import custom_template, format_title


def boxplot(
    data_frame: pd.DataFrame,
    x_axis: str,
    y_axis: str,
    title: str,
    subtitle: str | None = None,
    outliers: bool = True,
    **kwargs,
) -> go.Figure:
    """Create a box plot.

    Args:
        data_frame (pd.DataFrame): DataFrame to plot.
        x_axis (str): Column to use as x axis.
        y_axis (str): Column to use as y axis.
        title (str): Title of the plot.
        subtitle (str, optional): Subtitle of the plot. Defaults to None.
        outliers (bool, optional): Show outliers. Defaults to True.
        **kwargs: Keyword arguments to pass to plotly.graph_objects.Box.

    Returns:
        go.Figure: Box plot figure.
    """
    fig = go.Figure()
    for x_value in data_frame[x_axis].unique():
        y_values = data_frame[data_frame[x_axis] == x_value][y_axis]

        boxpoints = "outliers" if outliers else False

        fig.add_trace(
            go.Box(
                x=data_frame[data_frame[x_axis] == x_value][x_axis],
                y=y_values,
                name=str(x_value),
                boxpoints=boxpoints,
                **kwargs,
            )
        )

    fig.update_layout(
        title=format_title(title, subtitle=subtitle),
        template=custom_template,
        width=800,
        height=400,
        showlegend=False,
    )
    fig.update_xaxes(title_text=x_axis)
    fig.update_yaxes(title_text=y_axis)
    return fig



"""Missing values transformation."""
import pandas as pd


def fill_missing_values(
    data_frame: pd.DataFrame, columns: list[str], value: any
) -> pd.DataFrame:
    """Fill missing values in columns of a DataFrame.

    Args:
        data_frame (pd.DataFrame): DataFrame to fill missing values in.
        columns (list[str]): List of columns to fill missing values in.
        value (any): Value to fill missing values with.

    Returns:
        pd.DataFrame: DataFrame with missing values filled.
    """
    data_frame[columns] = data_frame[columns].fillna(value)
    return data_frame



"""Geo plot module."""
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from .style import custom_template, format_title


def geoplot(
    data_frame: pd.DataFrame,
    lat: str,
    lon: str,
    title: str,
    subtitle: str | None = None,
    **kwargs,
) -> go.Figure:
    """Create a map plot based on latitude and longitude of DataFrame passed columns.

    Args:
        data_frame (pd.DataFrame): DataFrame to plot.
        lat (str): Column to use as latitude.
        lon (str): Column to use as longitude.
        title (str): Title of the plot.
        subtitle (str, optional): Subtitle of the plot. Defaults to None.
        **kwargs: Keyword arguments to pass to plotly.express.scatter_geo.

    Returns:
        go.Figure: Bar plot figure.
    """
    fig = px.scatter_geo(
        data_frame,
        lat=lat,
        lon=lon,
        title=format_title(title, subtitle=subtitle),
        template=custom_template,
        projection="natural earth",
        width=800,
        height=400,
        **kwargs,
    )
    return fig



"""Statistics module for coolpandas."""
from .effect import (
    cohens_d_from_data,
    cohens_d_from_mde_absolute,
    cohens_d_from_mde_relative,
    cohens_d_from_statistics,
    compute_bootstrap_ci_cohens_d,
    compute_parametric_ci_cohens_d,
    hedges_g,
    pooled_standard_deviation,
    two_sample_t_test_pooled,
)
from .ttest import TTestInd



"""Plot module for coolpandas."""
from .barplot import barplot
from .boxplot import boxplot
from .distplot import distplot
from .geoplot import geoplot
from .lineplot import lineplot
from .mapplot import mapplot



"""CoolPandas missing_values functions."""
import pandas as pd

from coolpandas.plot import barplot


def plot_missing_values(null_data_frame: pd.DataFrame) -> None:
    """Plot missing values statistics for a DataFrame.

    Args:
        data_frame (pd.DataFrame): DataFrame to plot missing values statistics.
    """
    fig = barplot(
        null_data_frame,
        x_axis="column",
        y_axis="percentage",
        title="Missing values",
        hover_data={
            "null_values": True,
        },
        labels={
            "column": "Column",
            "null_values": "Missing values",
            "percentage": "Percentage",
        },
        subtitle="Missing values per column in DataFrame.",
        text="null_values",
    )
    fig.update_layout(yaxis={"ticksuffix": "%", "range": [0, 100]})
    fig.show()


def get_missing_values(data_frame: pd.DataFrame, plot: bool = True) -> pd.DataFrame:
    """Get columns missing values statistics for a DataFrame.

    Args:
        data_frame (pd.DataFrame): DataFrame to get missing values statistics.
        plot (bool, optional): Whether to plot the missing values statistics.
        Defaults to True.

    Returns:
        pd.DataFrame: Missing values statistics as DataFrame.
    """
    null_values: pd.Series = data_frame.isna().sum()
    null_data_frame: pd.DataFrame = (
        null_values.reset_index()
        .rename(columns={"index": "column", 0: "null_values"})
        .sort_values(by="null_values", ascending=False)
    )
    null_data_frame["percentage"] = round(
        null_data_frame["null_values"] / data_frame.shape[0] * 100, 2
    )
    null_data_frame = null_data_frame[null_data_frame["null_values"] > 0]
    null_data_frame = null_data_frame.reset_index(drop=True)
    print(f"Number of columns with missing values: {null_data_frame.shape[0]}")
    if plot:
        plot_missing_values(null_data_frame)
    return null_data_frame



import pandas as pd
import redshift_connector
from redshift_connector import Connection, Cursor


def get_redshift_cursor(host: str, database: str, user: str, password: str) -> Cursor:
    """Get a cursor to the Redshift database.

    Args:
        host (str): The host of the Redshift database
        database (str): The name of the database
        user (str): The user to connect to the database
        password (str): The password to connect to the database

    Returns:
        Cursor: A cursor to the Redshift database
    """
    conn: Connection = redshift_connector.connect(
        host=host, database=database, user=user, password=password
    )
    return conn.cursor()


def get_table_columns_information(cursor: Cursor, redshift_table: str) -> tuple:
    """Get the columns information of the table.

    Args:
        cursor (Cursor): A cursor to the Redshift database
        redshift_table (str): The name of the table

    Returns:
        tuple: A tuple containing the columns information of the table
    """
    cursor.execute(f"SELECT pg_get_cols('{redshift_table}');")
    return cursor.fetchall()


def query_to_dataframe(cursor: Cursor, query: str) -> pd.DataFrame:
    """Execute a query and return the result as a pandas DataFrame.

    Args:
        cursor (Cursor): A cursor to the Redshift database
        query (str): A query to execute

    Returns:
        pd.DataFrame: The result of the query as a pandas DataFrame
    """
    cursor.execute(query)
    return cursor.fetch_dataframe()
